#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "core.h"
#include "threads.h"
#include "sock.h"

#include "ocsp_request.h"

#define  OCSPD_DEF_MAX_SIZE 	65535
#define  OCSPD_DEF_MAX_READ	1024
#define  MAX_USEC		1000
#define  WAIT_USEC		50

extern OCSPD_CONFIG *ocspd_conf;

OCSP_REQUEST * ocspd_req_get_socket ( int connfd, OCSPD_CONFIG *ocspd_conf) {

	ssize_t fullsize = 0;
	ssize_t newsize  = 0;
	ssize_t maxsize  = 0;
	ssize_t	cont_len = 0;
	ssize_t full_req_size = 0;

	char*   buf      = NULL;

	int	post	 = 0;
	int	headers  = 0;
	int	sel_ret	 = 0;
	int	cont	 = 0;

	char    *pnt	 = NULL;
	char 	*pnt_end = NULL;
	char    *req_st  = NULL;

	BIO	*mem	 = NULL;
	OCSP_REQUEST * req = NULL;

	struct  timeval   time_out;
	fd_set  readset;

	if( !ocspd_conf ) {
		return (NULL);
	}

	if( ocspd_conf->max_req_size > 2 * OCSPD_DEF_MAX_READ ) {
		buf = (char *) malloc ( ocspd_conf->max_req_size );
		maxsize = ocspd_conf->max_req_size - OCSPD_DEF_MAX_READ;
	} else {
		buf = (char *) malloc ( OCSPD_DEF_MAX_SIZE );
		maxsize = OCSPD_DEF_MAX_SIZE - OCSPD_DEF_MAX_READ;
	}

	/* Add the socket to the read set */
	FD_ZERO( &readset );
	FD_SET (connfd, &readset);

	time_out.tv_sec = ocspd_conf->max_timeout_secs;
	time_out.tv_usec = 0;

	if (fcntl(connfd, F_SETFL, O_NONBLOCK) < 0) {
		syslog(LOG_ERR, "[Tread] ERROR, cannot set no non-waiting\n");
		free(buf);
		return(NULL);
	}

	cont = 0;
	/*
	while((sel_ret = Select(connfd+1, &readset, 
				NULL, NULL, &time_out)) >= 0) {
	*/
	full_req_size = 0;
	for(;;) {

		FD_ZERO( &readset );
		FD_SET (connfd, &readset);

		time_out.tv_sec = ocspd_conf->max_timeout_secs;
		time_out.tv_usec = 0;

		sel_ret = select(connfd+1, &readset, NULL, NULL, &time_out);

		if (sel_ret == -1 && errno == EINTR) {
			if(ocspd_conf->debug)
				syslog(LOG_ERR, "ERROR::NET::Select Recoverable %s",
					strerror(errno));
			continue;
		}

		if( sel_ret < 0 ) {
			if(ocspd_conf->debug )
				syslog(LOG_ERR, "ERROR::NET::Select %s",
					strerror(errno));

			free(buf);
			return(NULL);
		}

		if( sel_ret == 0 ) {
			if( ocspd_conf->debug )
				syslog(LOG_ERR, "ERROR::NET::Socket connection timeout");
			free(buf);
			return(NULL);
		}

		/*
		if( ocspd_conf->debug ) {
			syslog(LOG_ERR, "DEBUG::NETWORK::Select "
				"%d (cont = %d)!", sel_ret, cont);
		}
	
		if( ocspd_conf->debug ) {
			syslog(LOG_ERR, "DEBUG::NETWORK::FD_ISSET "
				"%d", FD_ISSET(connfd, &readset) );
		}
		*/
	
		// cont += WAIT_USEC;

		if (FD_ISSET (connfd, &readset)) {
			if((newsize = recv(connfd, &(buf[fullsize]), 
				OCSPD_DEF_MAX_READ, 0 )) == 0 ) {
				break;
			}

			if (newsize < 0) {
				if( ocspd_conf->debug ) {
					syslog(LOG_ERR, "DEBUG::NETWORK::ERROR in RECV::"
						"%s [%d]", strerror(errno), errno );
					if( errno == EWOULDBLOCK ) {
						syslog(LOG_ERR, "DEBUG::NETWORK::ERROR::"
							"EWOULDBLOCK");
						continue;
					}
				}
				break;
			}

			if( ocspd_conf->debug ) {
				syslog(LOG_ERR, "DEBUG::NETWORK::recv "
					"received %d (conn = %d)!", newsize, connfd);
			}
	
		/*
		if( sel_ret == 0 ) {
			if( (cont/1000) >= ocspd_conf->max_timeout_secs ) {
				if( ocspd_conf->verbose ) {
					syslog(LOG_ERR, "ERROR::NETWORK::Timeout "
					"reached while reading REQUEST (%d >= %d)!",
					cont/1000, ocspd_conf->max_timeout_secs);
				}

				break;
			} else {
				continue;
			}
		}
		*/

			/* Check for headers */
			if( (!cont_len) && 
				((pnt = strstr(buf, "Content-Length: " )) != NULL) ) {
				if((pnt_end = strstr(pnt, "\r\n")) != NULL )
					sscanf((char *) (pnt+16), "%d", &cont_len );
				if(ocspd_conf->debug) {
					syslog(LOG_ERR, "DEBUG::Got Content Len [%d]",
						cont_len);
				}
			}

			fullsize += newsize;
			if( (!headers) && 
				(((pnt = strstr(buf, "\r\n\r\n")) != NULL) ||
					((pnt = strstr(buf, "\n\n")) != NULL)) ) {

				while( (*pnt == '\n') || (*pnt == '\r')) {
					*pnt = '\x0';
					pnt++;
				}

				if( !cont_len ) {
					if(ocspd_conf->verbose) 
						syslog(LOG_ERR, "ERROR::No "
							"Content-Length"
							" in REQ Headers");
					if(ocspd_conf->debug) {
						fprintf( stderr,"---BEGIN HEADERS---\n"
							"%s\n"
							"---END HEADERS---\n\n",
							buf );
					}
					free(buf);
					return(NULL);
				}
				req_st = pnt;
				headers = (int) (req_st - buf);
				full_req_size = headers+cont_len;
			}

			if( (!post) && (fullsize >= 5 ) &&
				( strncmp( buf, "POST ", 5) != 0 )) {

				/* Got an error - probably not found (?) */
				if( ocspd_conf->verbose) {
					syslog( LOG_ERR, "ERROR::Request::HTTP method "
								"is not POST");
				}

				free(buf);
				return(NULL);
			}

			if( fullsize >= maxsize ) {
				/* Max Reading size exceeded */
				syslog( LOG_ERR, 
					"ERROR::Max REQUEST size exceeded [ %d ]",
						maxsize );
				free( buf );
				return(NULL);
			}

			if( (full_req_size > 0) && 
					(full_req_size - fullsize < 1 )) {

				break;
			}

		}
	}

	if(ocspd_conf->debug) {
		fprintf( stderr,"---BEGIN HEADERS---\n"
			"%s\n"
			"---END HEADERS---\n\n",
			buf );
	}

        if(!(mem = BIO_new_mem_buf(req_st, cont_len) )) {
		BIO *err = NULL;

		if((err = BIO_new(BIO_s_file())) != NULL) {
			BIO_set_fp(err,stderr,BIO_NOCLOSE);
			ERR_print_errors(err);
			BIO_free(err);
		}

                syslog( LOG_ERR, "ERROR: Internal memory allocation error!");
		if(ocspd_conf->debug) {
			fprintf(stderr, "ERROR::req_st=%p [len %d]\n",
				req_st, cont_len);
		}
        } else {
		if((req = d2i_OCSP_REQUEST_bio(mem, NULL)) == NULL ) {
			if(ocspd_conf->debug) {
				syslog(LOG_ERR, "ERROR::Parsing in d2i_ function\n");

				fprintf(stderr, "[len %d] buf=%p -- req_st=%p "
						"(fullsize %d - hd_size = %d - "
						"rq_size = %d)\n", 
						cont_len, buf, 
						req_st, fullsize,
						req_st - buf,
						fullsize - (req_st - buf) );
			}
		}
		if(mem) BIO_free (mem);
	}

	if( buf ) free (buf);

	return (req);
}

